/*
Stochastic Theory Education through Visualization Environment (STEVE) 2.0
Robert Florence, Tennessee Tech University
Window.cpp - This file creates the window and input form and enables OpenGL.
It also handles the Windows message loop for the main window and input form.
Last Modified: July 24, 2008
*/

#include <windows.h>
#include <gl/glut.h>
#include <cstdlib>
#include <ctime>
#include "resource.h"
#include "InputForm.h"
#include "Visualization.h"

//dimensions of window and form (pixels)
#define WND_WIDTH 1000
#define WND_HEIGHT 600
#define FORM_WIDTH 400
//animation modes
#define DRAW_BACK 700
#define DRAW_PAUSED 701
#define DRAW_FORWARD 702

//function declarations
void openDoc(char* filename);
void draw(int);
void resizeWnd(int, int);
void enableGL();
void disableGL();
BOOL CALLBACK WaitDlgProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK RepBugDlgProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK AckDlgProc(HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

//handles to main window and input form
HWND hWndMain = NULL;
HWND hWndInputForm = NULL;
//handles to device context and rendering context for OpenGL
HDC hDC = NULL;
HGLRC hRC = NULL;
//execution info for simulation program
SHELLEXECUTEINFO simExec;
//whether or not simulation program has been canceled
BOOL simCancel = FALSE;
//input form instance
InputForm *inputForm = 0;
//default and custom visualization instances
Visualization *defaultVis = 0, *customVis = 0;
//state of play/pause button
BOOL paused = FALSE;
//number of clock ticks to advance to next frame in animation
clock_t nextFrame = 0;

/*
open a file in the "doc" folder with its default program
filename: file to open
*/
void openDoc(char* filename)
{
    SHELLEXECUTEINFO docExec;
    
    //create shell execution info for document
    ZeroMemory(&docExec, sizeof(docExec));
    docExec.cbSize = sizeof(docExec);
    docExec.lpVerb = "open";
    docExec.lpFile = filename;
    docExec.lpDirectory = "doc";
    docExec.nShow = SW_SHOWNORMAL;
    
    //open document
    ShellExecuteEx(&docExec);
}

/*
display both visualizations (if applicable), map grids, and color bar
mode: DRAW_BACK, DRAW_PAUSED, or DRAW_FORWARD
*/
void draw(int mode)
{
    glClear(GL_COLOR_BUFFER_BIT);

    //display visualizations
    if (defaultVis)
    {
        //change frame number
        if (mode == DRAW_BACK)
            defaultVis->decFrame();
        else if (mode == DRAW_FORWARD)
            defaultVis->incFrame();

        //draw default visualization in map 1 location
        defaultVis->draw(MAP1_LEFT, MAP1_TOP);
    }
    if (customVis)
    {
        //change frame number
        if (mode == DRAW_BACK)
            customVis->decFrame();
        else if (mode == DRAW_FORWARD)
            customVis->incFrame();

        //draw custom visualization in map 2 location
        customVis->draw(MAP2_LEFT, MAP2_TOP);
    }

    //draw map grids and color bar
    Visualization::drawAxes(MAP1_LEFT, MAP1_TOP);
    Visualization::drawAxes(MAP2_LEFT, MAP2_TOP);
    Visualization::drawColorBar();

    SwapBuffers(hDC);

    //set time to draw next frame
    if (mode != DRAW_PAUSED)
        nextFrame = (clock_t) ((double) clock() + CLOCKS_PER_SEC*inputForm->getAnimDelay());
}

/*
adjust OpenGL display and input form size
w, h: new dimensions of the window (pixels)
*/
void resizeWnd(int w, int h)
{
    //extend input form to bottom of window
    MoveWindow(hWndInputForm, 0, 0, FORM_WIDTH, h, TRUE);
	
	//avoid division by zero
    if (h == 0)
        h = 1;

    //resize OpenGL display area
    glViewport(FORM_WIDTH, 0, w - FORM_WIDTH, h);

    //set orthographic projection
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1.6, 1.6, -1.6, 1.6, -100.0, 100.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

/*
enables OpenGL for the main window
*/
void enableGL()
{
    PIXELFORMATDESCRIPTOR pfd;
    int pxFormat;

    //create a pixel format descriptor
    ZeroMemory(&pfd, sizeof(pfd));
    pfd.nSize = sizeof(pfd); //size of structure
    pfd.nVersion = 1; //version number
    pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER; //pixel format
    pfd.iPixelType = PFD_TYPE_RGBA; //rgba format
    pfd.cColorBits = 16; //color depth
    pfd.cDepthBits = 16; //depth buffer
    pfd.iLayerType = PFD_MAIN_PLANE; //main drawing layer
    
    //get a device context
    if (!(hDC = GetDC(hWndMain)))
    {
        disableGL();
        MessageBox(NULL, "Device Context Creation Failed!", "Error", MB_ICONEXCLAMATION | MB_OK);
    }
    
    //find a matching pixel format
    if (!(pxFormat = ChoosePixelFormat(hDC, &pfd)))
    {
        disableGL();
        MessageBox(NULL, "Pixel Format Selection Failed!", "Error", MB_ICONEXCLAMATION | MB_OK);
    }

    //set the pixel format
    if (!SetPixelFormat(hDC, pxFormat, &pfd))
    {
        disableGL();
        MessageBox(NULL, "Set Pixel Format Failed!", "Error", MB_ICONEXCLAMATION | MB_OK);
    }

    //create a rendering context
    if (!(hRC = wglCreateContext(hDC)))
    {
        disableGL();
        MessageBox(NULL, "Rendering Context Creation Failed!", "Error", MB_ICONEXCLAMATION | MB_OK);
    }

    //activate the rendering context
    if (!wglMakeCurrent(hDC, hRC))
    {
        disableGL();
        MessageBox(NULL, "Rendering Context Activation Failed!", "Error", MB_ICONEXCLAMATION | MB_OK);
    }

    //initialize OpenGL settings
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
    resizeWnd(WND_WIDTH, WND_HEIGHT);
}

/*
disables OpenGl for the main window
*/
void disableGL()
{
    if (hRC)
    {
        //detach the RC from the DC
        if (!wglMakeCurrent(NULL, NULL))
            MessageBox(NULL, "Rendering Context Release Failed!", "Error", MB_ICONEXCLAMATION | MB_OK);

        //delete the RC
        if (!wglDeleteContext(hRC))
            MessageBox(NULL, "Rendering Context Deletion Failed!", "Error", MB_ICONEXCLAMATION | MB_OK);

        hRC = NULL;
    }
}

/*
Windows message handler for the running simulation dialog box
hWnd: handle to dialog box
msg: message to be handled
wParam, lParam: parameters for the message
return: TRUE if the message is handled, otherwise FALSE
*/
BOOL CALLBACK WaitDlgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch(msg)
    {
        //click a button
        case WM_COMMAND:
            switch(LOWORD(wParam))
            {
                //close dialog box
                case IDCANCEL:
                    //cancel simulation
                    simCancel = TRUE;
                    EndDialog(hWnd, IDCANCEL);
                break;
            }
        break;
        default:
            return FALSE;
    }

    return TRUE;
}

/*
Windows message handler for the report a bug dialog box
hWnd: handle to dialog box
msg: message to be handled
wParam, lParam: parameters for the message
return: TRUE if the message is handled, otherwise FALSE
*/
BOOL CALLBACK RepBugDlgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch(msg)
    {
        //click a button
        case WM_COMMAND:
            switch(LOWORD(wParam))
            {
                //close dialog box
                case IDCANCEL:
                    EndDialog(hWnd, IDCANCEL);
                break;
            }
        break;
        default:
            return FALSE;
    }

    return TRUE;
}

/*
Windows message handler for the acknowledgements dialog box
hWnd: handle to dialog box
msg: message to be handled
wParam, lParam: parameters for the message
return: TRUE if the message is handled, otherwise FALSE
*/
BOOL CALLBACK AckDlgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch(msg)
    {
        //create the dialog
        case WM_INITDIALOG:
            //load small TTU logo for acknowledgements dialog
            SendDlgItemMessage(hWnd, IDC_TTULOGOSM, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM) LoadImage(GetModuleHandle(NULL), "img/TTUlogoSm.bmp", IMAGE_BITMAP, 88, 65, LR_LOADFROMFILE | LR_LOADMAP3DCOLORS));
        break;
        //click a button
        case WM_COMMAND:
            switch(LOWORD(wParam))
            {
                //click url in acknowledgements dialog
                case IDC_ACKURL:
                    ShellExecute(hWnd, "open", "http://iweb.tntech.edu/saswe", NULL, NULL, SW_SHOWNORMAL);
                break;
                //close dialog box
                case IDCANCEL:
                    EndDialog(hWnd, IDCANCEL);
                break;
            }
        break;
        default:
            return FALSE;
    }
    
    return TRUE;
}

/*
Windows message handler for the main window and input form
hWnd: handle to dialog box
msg: message to be handled
wParam, lParam: parameters for the message
return: 0 if the message is handled, otherwise default window procedure
*/
LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch(msg)
    {
        //click a button or menu item
        case WM_COMMAND:
            switch(LOWORD(wParam))
            {
                //start button
                case IDC_START:
                    //check for valid inputs
                    if (inputForm->checkInputs())
                    {
                        //handle to running simulation dialog
                        HWND hWndWait;
                        //value of simulation period parameter
                        char simPeriodStr[10];
                        int simPeriod;
                        //start and end times of simulation execution
                        time_t start, end;
                        //message for running simulation message loop
                        MSG newMsg;
                        
                        //write parameters to simulation program's input file
                        inputForm->writeInputs("simul/params.dat");
                        
                        //get simulation period input value
                        GetDlgItemText(hWndInputForm, IDC_INPUTVAL + inputForm->getNumParams() - 1, simPeriodStr, 10);
                        simPeriod = atoi(simPeriodStr);

                        //create running simulation dialog
                        if (!(hWndWait = CreateDialog(GetModuleHandle(NULL),
                            MAKEINTRESOURCE(IDD_RUNSIM),
                            hWndMain,
                            WaitDlgProc
                            )))
                        {
                            MessageBox(NULL, "Waiting Dialog Box Creation Failed!", "Error", MB_ICONEXCLAMATION | MB_OK);
                        }
                        
                        //execute simulation program
                        time(&start);
                        if (ShellExecuteEx(&simExec))
                        {
                            simCancel = FALSE;
                            
                            //wait until simulation is finished
                            do
                            {
                                //read any messages in the message queue
                                PeekMessage(&newMsg, NULL, 0, 0, PM_REMOVE);
                                //only handle messages for running simulation dialog
                                IsDialogMessage(hWndWait, &newMsg);

                                //simulation has been canceled
                                if (simCancel)
                                {
                                    //terminate simulation program
                                    TerminateProcess(simExec.hProcess, 0);
                                    //destroy running simulation dialog
                                    DestroyWindow(hWndWait);

                                    return 0;
                                }
                            }
                            //continue loop if there is a message in the queue
                            //quit loop if simulation program finishes
                            while (MsgWaitForMultipleObjects(1, &simExec.hProcess, FALSE, INFINITE, QS_ALLINPUT) == WAIT_OBJECT_0 + 1);

                            //display simulation generation time
                            time(&end);
                            inputForm->dispGenTime((int) difftime(end, start));

                            //delete old visualizations
                            if (defaultVis)
                                delete defaultVis;
                            if (customVis)
                                delete customVis;

                            //create new visualizations
                            defaultVis = new Visualization(simPeriod, "simul/default.dat");
                            customVis = new Visualization(simPeriod, "simul/output.dat");

                            //update last simulation file
                            inputForm->writeInputs("simul/last.dat");
                            
                            //destroy running simulation dialog
                            DestroyWindow(hWndWait);

                            //draw first frame of visualizations
                            draw(DRAW_FORWARD);
                        }
                        //simulation execution failed
                        else
                        {
                            //destroy running simulation dialog
                            DestroyWindow(hWndWait);
                            
                            MessageBox(NULL, "Simulation Failed!", "Error", MB_ICONEXCLAMATION | MB_OK);
                        }
                    }
                break;
                //back button
                case IDC_BACK:
                    draw(DRAW_BACK);
                break;
                //play/pause button
                case IDC_PLAYPAUSE:
                    paused = inputForm->playPause(paused);
                break;
                //forward button
                case IDC_FORWARD:
                    draw(DRAW_FORWARD);
                break;
                //load settings menu item
                case ID_FILE_LOAD:
                    inputForm->loadSettings();
                break;
                //save settings menu item
                case ID_FILE_SAVE:
                    inputForm->saveSettings();
                break;
                //reset form menu item
                case ID_FILE_RESET:
                    inputForm->reset();
                break;
                //run last simulation menu item
                case ID_FILE_RUNLAST:
                    {
                        //value of simulation period parameter
                        char simPeriodStr[10];
                        int simPeriod;
                        
                        //set input form to values from last simulation
                        inputForm->readInputs("simul/last.dat");

                        //get simulation period input value
                        GetDlgItemText(hWndInputForm, IDC_INPUTVAL + inputForm->getNumParams() - 1, simPeriodStr, 10);
                        simPeriod = atoi(simPeriodStr);

                        //delete old visualizations
                        if (defaultVis)
                            delete defaultVis;
                        if (customVis)
                            delete customVis;

                        //create new visualizations
                        defaultVis = new Visualization(simPeriod, "simul/default.dat");
                        customVis = new Visualization(simPeriod, "simul/output.dat");

                        //draw first frame of visualizations
                        draw(DRAW_FORWARD);
                    }
                break;
                //report bug menu item
                case ID_FILE_REPBUG:
                    //create report bug dialog box
                    if (!DialogBox(GetModuleHandle(NULL),
                        MAKEINTRESOURCE(IDD_REPBUG),
                        hWndMain,
                        RepBugDlgProc
                        ))
                    {
                        MessageBox(NULL, "Report a Bug Dialog Box Creation Failed!", "Error", MB_ICONEXCLAMATION | MB_OK);
                    }
                break;
                //exit menu item
                case ID_FILE_EXIT:
                    PostMessage(hWnd, WM_CLOSE, 0, 0);
                break;
                //about steve menu item
                case ID_HELP_ABOUT:
                    openDoc("Introduction-STEVE.pdf");
                break;
                //steve user guide menu item
                case ID_HELP_USERGUIDE:
                    openDoc("STEVE-User_Guide.pdf");
                break;
                //SREM2D theory menu item
                case ID_HELP_SREM2D:
                    openDoc("SREM2D_Theory.pdf");
                break;
                //general information on metrics menu item
                case ID_HELP_GIM:
                    openDoc("General_User_Document.pdf");
                break;
                //acknowledgements menu item
                case ID_HELP_ACK:
                    //create acknowledgements dialog box
                    if (!DialogBox(GetModuleHandle(NULL),
                        MAKEINTRESOURCE(IDD_ACK),
                        hWndMain,
                        AckDlgProc
                        ))
                    {
                        MessageBox(NULL, "Acknowledgements Dialog Box Creation Failed!", "Error", MB_ICONEXCLAMATION | MB_OK);
                    }
                break;
            }
        break;
        //resize the window
        case WM_SIZE:
            resizeWnd(LOWORD(lParam), HIWORD(lParam));
        break;
        //close the window
        case WM_CLOSE:
            DestroyWindow(hWnd);
        break;
        //exit the program
        case WM_DESTROY:
            PostQuitMessage(0);
        break;
        default:
            return DefWindowProc(hWnd, msg, wParam, lParam);
    }

    return 0;
}

/*
create window, input form, and simulation execution info,
    and start Windows message loop
hInstance: handle to program instance
hPrevInstance, lpCmdLine, nCmdShow: not used
*/
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
    //window class
    WNDCLASSEX wc;
    //window class name
    char szClassName[] = "STEVE";
    //rectangle for viewable area of window
    RECT wndRect;
    //message retrieved from message loop
    MSG msg;
    //whether or not message loop is finished
    BOOL done = FALSE;

    //create the window class
    wc.hInstance = hInstance; //application instance handle
    wc.lpszClassName = szClassName; //class name
    wc.lpfnWndProc = WndProc; //window message procedure
    wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC; //redraw when resized
    wc.cbSize = sizeof(WNDCLASSEX); //size of the structure
    wc.hIcon = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_ICON)); //large icon
    wc.hIconSm = (HICON) LoadImage(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_ICON), IMAGE_ICON, 16, 16, 0); //small icon
    wc.hCursor = LoadCursor(NULL, IDC_ARROW); //default cursor
    wc.lpszMenuName = MAKEINTRESOURCE(IDR_MENU); //menu handle
    wc.cbClsExtra = 0; //no extra allocated memory
    wc.cbWndExtra = 0; //no extra allocated memory for windows of this type
    wc.hbrBackground = (HBRUSH) COLOR_3DSHADOW; //background color

    //register the window class
    if(!RegisterClassEx(&wc))
    {
        MessageBox(NULL, "Window Registration Failed!", "Error", MB_ICONEXCLAMATION | MB_OK);
        return 0;
    }

    //create window rectangle
    SetRect(&wndRect, 0, 0, WND_WIDTH, WND_HEIGHT);
    AdjustWindowRect(&wndRect, WS_OVERLAPPEDWINDOW, TRUE);

    //create the main window
    if (!(hWndMain = CreateWindow(
        szClassName, //class name
        "S.T.E.V.E. 2.0 - Stochastic Theory Education through Visualization Environment", //window title
        WS_OVERLAPPEDWINDOW | WS_VISIBLE | WS_CLIPCHILDREN | WS_CLIPSIBLINGS, //window style
        CW_USEDEFAULT, CW_USEDEFAULT, //window position (pixels)
        wndRect.right - wndRect.left, wndRect.bottom - wndRect.top, //window size (pixels)
        NULL, //no parent window
        NULL, //no menu
        hInstance, //application instance handle
        NULL //window creation data
        )))
    {
        MessageBox(NULL, "Window Creation Failed!", "Error", MB_ICONEXCLAMATION | MB_OK);
        return 0;
    }
    
    //create the input form window
    if (!(hWndInputForm = CreateWindow(
        szClassName, //class name
        "", //no window title
        WS_CHILD | WS_VISIBLE, //window styles
        0, 0, //window position (relative)
        FORM_WIDTH, wndRect.bottom - wndRect.top, //window size (pixels)
        hWndMain, //parent window
        NULL, //no menu
        hInstance, //application instance handle
        NULL //window creation data
        )))
    {
        MessageBox(NULL, "Input Window Creation Failed!", "Error", MB_ICONEXCLAMATION | MB_OK);
        return 0;
    }
    
    //create the input form
    inputForm = new InputForm(hWndInputForm, "simul/paramInfo.dat");
    if (!inputForm->createWindows())
    {
        MessageBox(NULL, "Input Form Creation Failed!", "Error", MB_ICONEXCLAMATION | MB_OK);
        return 0;
    }
    
    //create shell execution info for fortran program
    ZeroMemory(&simExec, sizeof(simExec));
    simExec.cbSize = sizeof(simExec);
    simExec.fMask = SEE_MASK_NOCLOSEPROCESS;
    simExec.lpVerb = "open";
    simExec.lpFile = "simulation_fast.exe";
    simExec.lpDirectory = "simul";
    simExec.nShow = SW_HIDE;
    
    enableGL();

    //message loop
    while (!done)
    {
        //handle messages
        if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
        {
            //exit program
            if (msg.message == WM_QUIT)
                done = TRUE;
            //input form messages
            else if (!IsDialogMessage(hWndInputForm, &msg))
            {
                //other messages
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            }
        }
        //advance and draw frame
        else if (!paused && clock() >= nextFrame)
            draw(DRAW_FORWARD);
        //draw frame
        else
            draw(DRAW_PAUSED);
    }

    disableGL();

    return msg.wParam;
}
